# -*- encoding: utf-8 -*-
"""
tests.core.didding module

"""

import os
import sys
from unittest.mock import patch

import pytest
from hio.help.hicting import Mict
from keri.app import habbing
from keri.core import coring, eventing, serdering
from keri.db import basing, koming, subing
from keri.vdr import credentialing, verifying
from mockito import mock, unstub, when

from dws import DidWebsError, UnknownAID
from dws.core import didding, didkeri

sys.path.append(os.path.join(os.path.dirname(__file__)))

wdid = 'did:webs:127.0.0.1:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
did = 'did:webs:127.0.0.1:ECCoHcHP1jTAW8Dr44rI2kWzfF71_U0sZwvV-J_q4YE7'


def test_parse_keri_did():
    # Valid did:keri DID
    did = 'did:keri:EKW4IEkAZ8VQ_ADXbtRsSOQ_Gk0cRxp6U4qKSr4Eb8zg'
    aid, query = didding.parse_did_keri(did)
    assert aid == 'EKW4IEkAZ8VQ_ADXbtRsSOQ_Gk0cRxp6U4qKSr4Eb8zg'

    # Invalid AID in did:keri
    did = 'did:keri:Gk0cRxp6U4qKSr4Eb8zg'

    with pytest.raises(ValueError) as e:
        _, _ = didding.parse_did_keri(did)

    assert isinstance(e.value, ValueError)
    assert str(e.value) == 'Gk0cRxp6U4qKSr4Eb8zg is an invalid AID'

    non_matching_dids = [
        'did:keri:example:extra',
        'did:keri:',
        'did:keri:example:123',
        'did:keri:example:extra:more',
        'did:keri:example:extra:evenmore',
    ]

    for did in non_matching_dids:
        with pytest.raises(ValueError):
            didding.parse_did_keri(did)

        assert isinstance(e.value, ValueError)

    with pytest.raises(ValueError) as e:
        didding.generate_did_doc(hby=mock(), rgy=mock(), did='did:keri:', aid='EKW4IEkAZ8VQ_ADXbtRsSOQ_Gk0cRxp6U4qKSr4Eb8zg')


def test_parse_webs_did():
    with pytest.raises(ValueError) as e:
        did = 'did:example'
        didding.parse_did_webs(did)
    assert str(e.value) == f'{did} is not a valid did:web(s) DID'

    with pytest.raises(ValueError) as e:
        did = 'did:webs:127.0.0.1:1234567'
        domain, port, path, aid, query = didding.parse_did_webs(did)

    assert isinstance(e.value, ValueError)
    assert str(e.value) == '1234567 is an invalid AID'

    did = 'did:webs:127.0.0.1:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did)
    assert '127.0.0.1' == domain
    assert None == port
    assert None == path
    assert aid == 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'

    # port url should be url encoded with %3a according to the spec
    did_port_bad = 'did:webs:127.0.0.1:7676:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did_port_bad)
    assert '127.0.0.1' == domain
    assert None == port
    assert '7676' == path
    assert aid == 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'

    did_port = 'did:webs:127.0.0.1%3a7676:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did_port)
    assert '127.0.0.1' == domain
    assert '7676' == port
    assert None == path
    assert aid == 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'

    # port should be url encoded with %3a according to the spec
    did_port_path_bad = 'did:webs:127.0.0.1:7676:my:path:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did_port_path_bad)
    assert '127.0.0.1' == domain
    assert None == port
    assert '7676:my:path' == path
    assert aid == 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'

    # port is properly url encoded with %3a according to the spec
    did_port_path = 'did:webs:127.0.0.1%3a7676:my:path:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did_port_path)
    assert '127.0.0.1' == domain
    assert '7676' == port
    assert 'my:path' == path
    assert aid == 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'

    did_path = 'did:webs:127.0.0.1:my:path:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did_path)
    assert '127.0.0.1' == domain
    assert None == port
    assert 'my:path' == path
    assert aid, 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'


def test_parse_web_did():
    with pytest.raises(ValueError) as e:
        did = 'did:web:127.0.0.1:1234567'
        domain, port, path, aid, query = didding.parse_did_webs(did)

    assert isinstance(e.value, ValueError)

    did = 'did:web:127.0.0.1:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did)
    assert '127.0.0.1' == domain
    assert None == port
    assert None == path
    assert aid == 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'

    # port url should be url encoded with %3a according to the spec
    did_port_bad = 'did:web:127.0.0.1:7676:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did_port_bad)
    assert '127.0.0.1' == domain
    assert None == port
    assert '7676' == path
    assert aid == 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'

    did_port = 'did:web:127.0.0.1%3a7676:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did_port)
    assert '127.0.0.1' == domain
    assert '7676' == port
    assert None == path
    assert aid == 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'

    # port should be url encoded with %3a according to the spec
    did_port_path_bad = 'did:web:127.0.0.1:7676:my:path:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did_port_path_bad)
    assert '127.0.0.1' == domain
    assert None == port
    assert '7676:my:path' == path
    assert aid == 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'

    # port is properly url encoded with %3a according to the spec
    did_port_path = 'did:web:127.0.0.1%3a7676:my:path:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did_port_path)
    assert '127.0.0.1' == domain
    assert '7676' == port
    assert 'my:path' == path
    assert aid == 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'

    did_path = 'did:web:127.0.0.1:my:path:BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'
    domain, port, path, aid, query = didding.parse_did_webs(did_path)
    assert '127.0.0.1' == domain
    assert None == port
    assert 'my:path' == path
    assert aid, 'BBilc4-L3tFUnfM_wJr4S4OJanAv_VmF_dJNN6vkf2Ha'


def test_generate_did_doc_bad_aid():
    hby = mock()
    rgy = mock()
    did = 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    aid = 'EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HD'

    with pytest.raises(ValueError) as e:
        didding.generate_did_doc(hby=hby, rgy=rgy, did=did, aid=aid)

    assert isinstance(e.value, ValueError)
    assert str(e.value) == (
        'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4 does '
        'not contain AID EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HD'
    )


def test_generate_did_doc_unknown_aid():
    hby = mock()
    hab = mock()
    hab_db = mock()
    kever = mock()
    db = mock()
    roobi = mock()
    rgy = mock()

    did = 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    aid = 'EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    hab.name = 'test_hab'
    hab.db = hab_db
    hby.habs = {aid: hab}
    db.roobi = roobi
    hby.db = db
    hby.kevers = {'a different aid': kever}

    with pytest.raises(UnknownAID) as e:
        didding.generate_did_doc(hby=hby, rgy=rgy, did=did, aid=aid)

    assert isinstance(e.value, UnknownAID)
    assert (
        str(e.value)
        == 'Unknown AID EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4 found in did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    )


def role_urls_fixture():
    return Mict(
        [
            (
                'controller',
                Mict(
                    [
                        (
                            'BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX',
                            Mict(
                                [
                                    ('http', 'http://localhost:8080/witness/wok'),
                                    ('tcp', 'tcp://localhost:8080/witness/wok'),
                                ]
                            ),
                        )
                    ]
                ),
            )
        ]
    )


def witness_urls_fixture():
    return Mict(
        [
            (
                'witness',
                Mict(
                    [
                        (
                            'BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX',
                            Mict([('http', 'http://localhost:8080/witness/wok')]),
                        )
                    ]
                ),
            )
        ]
    )


def test_generate_did_doc_single_sig():
    hby = mock()
    hby.name = 'test_hby'
    hab = mock()
    hab_db = mock()
    kever = mock()
    verfer = mock()
    tholder = mock()
    db = mock()
    locs = mock()

    did = 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    aid = 'EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    hab.name = 'test_hab'
    hab.db = hab_db
    hab.delpre = None
    hby.habs = {aid: hab}
    sner = mock()
    sner.num = 0
    kever.sner = sner
    hby.kevers = {aid: kever}
    verfer.raw = bytearray()
    verfer.qb64 = 'DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K'
    kever.verfers = [verfer]
    tholder.thold = None
    kever.tholder = tholder
    db.locs = locs
    hby.db = db
    wits = []
    kever.wits = wits

    loc = basing.LocationRecord(url='tcp://127.0.0.1:5634/')
    when(db.locs).getItemIter(keys=(aid,)).thenReturn([((aid, 'some_key'), loc)])

    when(hab).fetchRoleUrls(cid=aid).thenReturn(role_urls_fixture())
    when(hab).fetchWitnessUrls(cid=aid).thenReturn(witness_urls_fixture())

    rgy = mock()
    issus = mock()
    schms = mock()
    rgy.reger = mock()
    rgy.reger.issus = issus
    rgy.reger.schms = schms
    vry = mock()

    when(credentialing).Regery(hby=hby, name=hby.name).thenReturn(rgy)
    when(verifying).Verifier(hby=hby, reger=rgy.reger).thenReturn(vry)

    when(rgy.reger.issus).get(keys=aid).thenReturn([])
    when(rgy.reger.issus).get(keys=aid).thenReturn([])

    when(rgy.reger).cloneCreds([], hab_db).thenReturn([])

    diddoc = didding.generate_did_doc(hby=hby, rgy=rgy, did=did, aid=aid)

    assert diddoc == {
        'id': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
        'verificationMethod': [
            {
                'id': '#DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                'type': 'JsonWebKey',
                'controller': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                'publicKeyJwk': {
                    'kid': 'DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                    'kty': 'OKP',
                    'crv': 'Ed25519',
                    'x': '',
                },
            }
        ],
        'service': [
            {
                'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/controller',
                'serviceEndpoint': {'http': 'http://localhost:8080/witness/wok', 'tcp': 'tcp://localhost:8080/witness/wok'},
                'type': 'controller',
            },
            {
                'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/witness',
                'serviceEndpoint': {'http': 'http://localhost:8080/witness/wok'},
                'type': 'witness',
            },
        ],
        'alsoKnownAs': [],
    }

    unstub()


def test_generate_did_doc_single_sig_with_designated_alias(mock_helping_now_utc):
    hby = mock()
    hby.name = 'test_hby'
    hab = mock()
    hab_db = mock()
    kever = mock()
    verfer = mock()
    tholder = mock()
    db = mock()
    locs = mock()

    did = 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    aid = 'EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    hab.name = 'test_hab'
    hab.db = hab_db
    hab.delpre = None
    hby.habs = {aid: hab}
    sner = mock()
    sner.num = 0
    kever.sner = sner
    hby.kevers = {aid: kever}
    verfer.raw = bytearray()
    verfer.qb64 = 'DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K'
    kever.verfers = [verfer]
    tholder.thold = None
    kever.tholder = tholder
    db.locs = locs
    hby.db = db
    wits = []
    kever.wits = wits

    loc = basing.LocationRecord(url='tcp://127.0.0.1:5634/')
    when(db.locs).getItemIter(keys=(aid,)).thenReturn([((aid, 'some_key'), loc)])

    when(hab).fetchRoleUrls(cid=aid).thenReturn(role_urls_fixture())
    when(hab).fetchWitnessUrls(cid=aid).thenReturn(witness_urls_fixture())

    rgy = mock()
    issus = mock()
    schms = mock()
    rgy.reger = mock()
    rgy.reger.issus = issus
    rgy.reger.schms = schms
    vry = mock()

    when(credentialing).Regery(hby=hby, name=hby.name).thenReturn(rgy)
    when(verifying).Verifier(hby=hby, reger=rgy.reger).thenReturn(vry)

    cred1 = mock({'qb64': 'cred_1_qb64'}, coring.Saider)
    cred2 = mock({'qb64': 'cred_2_qb64'}, coring.Saider)
    when(rgy.reger.issus).get(keys=aid).thenReturn([cred1, cred2])
    when(rgy.reger.schms).get(keys='EN6Oh5XSD5_q2Hgu-aqpdfbVepdpYpFlgz6zvJL5b_r5').thenReturn([cred1, cred2])

    cloned_cred1 = {'sad': {'a': {'ids': ['designated_id_1']}}, 'status': {'et': 'iss'}}
    cloned_cred2 = {
        'sad': {'a': {'ids': ['did:webs:foo:designated_id_2', 'designated_id_2_but_different']}},
        'status': {'et': 'bis'},
    }
    when(rgy.reger).cloneCreds([cred1, cred2], hab_db).thenReturn([cloned_cred1, cloned_cred2])

    diddoc = didding.generate_did_doc(hby=hby, rgy=rgy, did=did, aid=aid)
    assert diddoc == {
        'id': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
        'verificationMethod': [
            {
                'id': '#DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                'type': 'JsonWebKey',
                'controller': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                'publicKeyJwk': {
                    'kid': 'DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                    'kty': 'OKP',
                    'crv': 'Ed25519',
                    'x': '',
                },
            }
        ],
        'service': [
            {
                'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/controller',
                'type': 'controller',
                'serviceEndpoint': {'http': 'http://localhost:8080/witness/wok', 'tcp': 'tcp://localhost:8080/witness/wok'},
            },
            {
                'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/witness',
                'type': 'witness',
                'serviceEndpoint': {'http': 'http://localhost:8080/witness/wok'},
            },
        ],
        'alsoKnownAs': ['designated_id_1', 'did:webs:foo:designated_id_2', 'designated_id_2_but_different'],
    }

    unstub()

    loc = basing.LocationRecord(url='tcp://127.0.0.1:5634/')
    when(db.locs).getItemIter(keys=(aid,)).thenReturn([((aid, 'some_key'), loc)])

    when(hab).fetchRoleUrls(cid=aid).thenReturn(role_urls_fixture())
    when(hab).fetchWitnessUrls(cid=aid).thenReturn(witness_urls_fixture())
    when(credentialing).Regery(hby=hby, name=hby.name).thenReturn(rgy)
    when(verifying).Verifier(hby=hby, reger=rgy.reger).thenReturn(vry)

    cred1 = mock({'qb64': 'cred_1_qb64'}, coring.Saider)
    cred2 = mock({'qb64': 'cred_2_qb64'}, coring.Saider)
    when(rgy.reger.issus).get(keys=aid).thenReturn([cred1, cred2])
    when(rgy.reger.schms).get(keys='EN6Oh5XSD5_q2Hgu-aqpdfbVepdpYpFlgz6zvJL5b_r5').thenReturn([cred1, cred2])

    cloned_cred1 = {'sad': {'a': {'ids': ['designated_id_1']}}, 'status': {'et': 'iss'}}
    cloned_cred2 = {
        'sad': {'a': {'ids': ['did:webs:foo:designated_id_2', 'designated_id_2_but_different']}},
        'status': {'et': 'bis'},
    }
    when(rgy.reger).cloneCreds([cred1, cred2], hab_db).thenReturn([cloned_cred1, cloned_cred2])

    diddoc = didding.generate_did_doc(hby=hby, rgy=rgy, did=did, aid=aid, meta=True)
    assert diddoc == {
        'didDocument': {
            'id': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
            'verificationMethod': [
                {
                    'id': '#DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                    'type': 'JsonWebKey',
                    'controller': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                    'publicKeyJwk': {
                        'kid': 'DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                        'kty': 'OKP',
                        'crv': 'Ed25519',
                        'x': '',
                    },
                }
            ],
            'service': [
                {
                    'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/controller',
                    'type': 'controller',
                    'serviceEndpoint': {
                        'http': 'http://localhost:8080/witness/wok',
                        'tcp': 'tcp://localhost:8080/witness/wok',
                    },
                },
                {
                    'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/witness',
                    'type': 'witness',
                    'serviceEndpoint': {'http': 'http://localhost:8080/witness/wok'},
                },
            ],
            'alsoKnownAs': ['designated_id_1', 'did:webs:foo:designated_id_2', 'designated_id_2_but_different'],
        },
        'didResolutionMetadata': {'contentType': 'application/did+json', 'retrieved': '2021-01-01T00:00:00Z'},
        'didDocumentMetadata': {'witnesses': [], 'versionId': '0', 'equivalentId': ['did:webs:foo:designated_id_2']},
    }


def test_generate_did_doc_single_sig_meta(mock_helping_now_utc):
    hby = mock()
    hby.name = 'test_hby'
    hab = mock()
    hab_db = mock()
    kever = mock()
    verfer = mock()
    tholder = mock()
    db = mock()
    locs = mock()

    did = 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    aid = 'EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    hab.name = 'test_hab'
    hab.db = hab_db
    hab.delpre = None
    hby.habs = {aid: hab}
    sner = mock()
    sner.num = 0
    kever.sner = sner
    hby.kevers = {aid: kever}
    verfer.raw = bytearray()
    verfer.qb64 = 'DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K'
    kever.verfers = [verfer]
    tholder.thold = None
    kever.tholder = tholder
    db.locs = locs
    hby.db = db
    kever.wits = ['witness1', 'witness2']

    loc = basing.LocationRecord(url='tcp://127.0.0.1:5632/')
    when(db.locs).getItemIter(keys=('witness1',)).thenReturn([(('witness1', 'some_key_witness1'), loc)])
    loc = basing.LocationRecord(url='tcp://127.0.0.1:5633/')
    when(db.locs).getItemIter(keys=('witness2',)).thenReturn([(('witness2', 'some_key_witness2'), loc)])
    loc = basing.LocationRecord(url='tcp://127.0.0.1:5634/')
    when(db.locs).getItemIter(keys=(aid,)).thenReturn([((aid, 'some_key'), loc)])

    when(hab).fetchRoleUrls(cid=aid).thenReturn(role_urls_fixture())
    when(hab).fetchWitnessUrls(cid=aid).thenReturn(witness_urls_fixture())

    rgy = mock()
    issus = mock()
    schms = mock()
    rgy.reger = mock()
    rgy.reger.issus = issus
    rgy.reger.schms = schms
    vry = mock()

    when(credentialing).Regery(hby=hby, name=hby.name).thenReturn(rgy)
    when(verifying).Verifier(hby=hby, reger=rgy.reger).thenReturn(vry)

    when(rgy.reger.issus).get(keys=aid).thenReturn([])
    when(rgy.reger.issus).get(keys=aid).thenReturn([])

    when(rgy.reger).cloneCreds([], hab_db).thenReturn([])

    diddoc = didding.generate_did_doc(hby=hby, rgy=rgy, did=did, aid=aid, meta=True)

    assert diddoc == {
        'didDocument': {
            'id': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
            'verificationMethod': [
                {
                    'id': '#DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                    'type': 'JsonWebKey',
                    'controller': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                    'publicKeyJwk': {
                        'kid': 'DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                        'kty': 'OKP',
                        'crv': 'Ed25519',
                        'x': '',
                    },
                }
            ],
            'service': [
                {
                    'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/controller',
                    'type': 'controller',
                    'serviceEndpoint': {
                        'http': 'http://localhost:8080/witness/wok',
                        'tcp': 'tcp://localhost:8080/witness/wok',
                    },
                },
                {
                    'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/witness',
                    'type': 'witness',
                    'serviceEndpoint': {'http': 'http://localhost:8080/witness/wok'},
                },
            ],
            'alsoKnownAs': [],
        },
        'didResolutionMetadata': {'contentType': 'application/did+json', 'retrieved': '2021-01-01T00:00:00Z'},
        'didDocumentMetadata': {
            'witnesses': [
                {'idx': '0', 'scheme': 'some_key_witness1', 'url': 'tcp://127.0.0.1:5632/'},
                {'idx': '1', 'scheme': 'some_key_witness2', 'url': 'tcp://127.0.0.1:5633/'},
            ],
            'versionId': '0',
            'equivalentId': [],
        },
    }

    unstub()


def test_generate_did_doc_multi_sig():
    hby = mock()
    hby.name = 'test_hby'
    hab = mock()
    hab_db = mock()
    kever = mock()
    verfer = mock()
    verfer_multi = mock()
    tholder = mock()
    db = mock()
    locs = mock()

    did = 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    aid = 'EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    hab.name = 'test_hab'
    hab.delpre = None
    hab.db = hab_db
    hby.habs = {aid: hab}
    sner = mock()
    sner.num = 0
    kever.sner = sner
    hby.kevers = {aid: kever}
    verfer.raw = bytearray()
    verfer.qb64 = 'DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K'
    verfer_multi.raw = bytearray()
    verfer_multi.qb64 = 'DOZlWGPfDHLMf62zSFzE8thHmnQUOgA3_Y-KpOyF9ScG'
    kever.verfers = [verfer, verfer_multi]
    tholder.thold = 2
    kever.tholder = tholder
    db.locs = locs
    hby.db = db
    wits = []
    kever.wits = wits

    loc = basing.LocationRecord(url='tcp://127.0.0.1:5634/')
    when(db.locs).getItemIter(keys=(aid,)).thenReturn([((aid, 'some_key'), loc)])

    when(hab).fetchRoleUrls(cid=aid).thenReturn(role_urls_fixture())
    when(hab).fetchWitnessUrls(cid=aid).thenReturn(witness_urls_fixture())

    rgy = mock()
    issus = mock()
    schms = mock()
    rgy.reger = mock()
    rgy.reger.issus = issus
    rgy.reger.schms = schms
    vry = mock()

    when(credentialing).Regery(hby=hby, name=hby.name).thenReturn(rgy)
    when(verifying).Verifier(hby=hby, reger=rgy.reger).thenReturn(vry)

    when(rgy.reger.issus).get(keys=aid).thenReturn([])
    when(rgy.reger.issus).get(keys=aid).thenReturn([])

    when(rgy.reger).cloneCreds([], hab_db).thenReturn([])

    diddoc = didding.generate_did_doc(hby=hby, rgy=rgy, did=did, aid=aid)

    assert diddoc == {
        'id': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
        'verificationMethod': [
            {
                'id': '#DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                'type': 'JsonWebKey',
                'controller': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                'publicKeyJwk': {
                    'kid': 'DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                    'kty': 'OKP',
                    'crv': 'Ed25519',
                    'x': '',
                },
            },
            {
                'id': '#DOZlWGPfDHLMf62zSFzE8thHmnQUOgA3_Y-KpOyF9ScG',
                'type': 'JsonWebKey',
                'controller': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                'publicKeyJwk': {
                    'kid': 'DOZlWGPfDHLMf62zSFzE8thHmnQUOgA3_Y-KpOyF9ScG',
                    'kty': 'OKP',
                    'crv': 'Ed25519',
                    'x': '',
                },
            },
            {
                'id': '#EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                'type': 'ConditionalProof2022',
                'controller': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                'threshold': 2,
                'conditionThreshold': [
                    '#DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                    '#DOZlWGPfDHLMf62zSFzE8thHmnQUOgA3_Y-KpOyF9ScG',
                ],
            },
        ],
        'service': [
            {
                'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/controller',
                'type': 'controller',
                'serviceEndpoint': {'http': 'http://localhost:8080/witness/wok', 'tcp': 'tcp://localhost:8080/witness/wok'},
            },
            {
                'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/witness',
                'type': 'witness',
                'serviceEndpoint': {'http': 'http://localhost:8080/witness/wok'},
            },
        ],
        'alsoKnownAs': [],
    }

    unstub()

    kever.tholder = coring.Tholder(sith=['1/2', '1/2'])

    when(db.locs).getItemIter(keys=(aid,)).thenReturn([((aid, 'some_key'), loc)])

    when(hab).fetchRoleUrls(cid=aid).thenReturn(role_urls_fixture())
    when(hab).fetchWitnessUrls(cid=aid).thenReturn(witness_urls_fixture())
    when(credentialing).Regery(hby=hby, name=hby.name).thenReturn(rgy)
    when(verifying).Verifier(hby=hby, reger=rgy.reger).thenReturn(vry)

    when(rgy.reger.issus).get(keys=aid).thenReturn([])
    when(rgy.reger.issus).get(keys=aid).thenReturn([])

    when(rgy.reger).cloneCreds([], hab_db).thenReturn([])

    diddoc = didding.generate_did_doc(hby=hby, rgy=rgy, did=did, aid=aid)

    assert diddoc == {
        'id': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
        'verificationMethod': [
            {
                'id': '#DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                'type': 'JsonWebKey',
                'controller': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                'publicKeyJwk': {
                    'kid': 'DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K',
                    'kty': 'OKP',
                    'crv': 'Ed25519',
                    'x': '',
                },
            },
            {
                'id': '#DOZlWGPfDHLMf62zSFzE8thHmnQUOgA3_Y-KpOyF9ScG',
                'type': 'JsonWebKey',
                'controller': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                'publicKeyJwk': {
                    'kid': 'DOZlWGPfDHLMf62zSFzE8thHmnQUOgA3_Y-KpOyF9ScG',
                    'kty': 'OKP',
                    'crv': 'Ed25519',
                    'x': '',
                },
            },
            {
                'id': '#EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                'type': 'ConditionalProof2022',
                'controller': 'did:web:127.0.0.1%3A7676:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
                'threshold': 1.0,
                'conditionWeightedThreshold': [
                    {'condition': '#DHGb2qY9WwZ1sBnC9Ip0F-M8QjTM27ftI-3jTGF9mc6K', 'weight': 1},
                    {'condition': '#DOZlWGPfDHLMf62zSFzE8thHmnQUOgA3_Y-KpOyF9ScG', 'weight': 1},
                ],
            },
        ],
        'service': [
            {
                'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/controller',
                'type': 'controller',
                'serviceEndpoint': {'http': 'http://localhost:8080/witness/wok', 'tcp': 'tcp://localhost:8080/witness/wok'},
            },
            {
                'id': '#BKVb58uITf48YoMPz8SBOTVwLgTO9BY4oEXRPoYIOErX/witness',
                'type': 'witness',
                'serviceEndpoint': {'http': 'http://localhost:8080/witness/wok'},
            },
        ],
        'alsoKnownAs': [],
    }

    unstub()


def test_to_did_web():
    diddoc = {'id': 'did:webs:example:123', 'verificationMethod': [{'controller': 'did:webs:example:123'}]}

    from dws.core.didding import to_did_web

    result = to_did_web(diddoc)

    assert result['id'] == 'did:web:example:123'
    assert result['verificationMethod'][0]['controller'] == 'did:web:example:123'

    with pytest.raises(DidWebsError) as excinfo:
        to_did_web(None)
    assert str(excinfo.value) == 'Cannot convert empty diddoc to did:web'

    unstub()


def test_from_did_web():
    diddoc = {'id': 'did:web:example:123', 'verificationMethod': [{'controller': 'did:web:example:123'}]}

    from dws.core.didding import from_did_web

    result = from_did_web(diddoc)

    # Verify the changes
    assert result['id'] == 'did:webs:example:123'
    assert result['verificationMethod'][0]['controller'] == 'did:webs:example:123'

    with pytest.raises(ValueError) as excinfo:
        from_did_web(diddoc, meta=True)
    assert str(excinfo.value) == f"Expected '{didding.DD_FIELD}' in did.json when indicating resolution metadata in use."

    unstub()


def test_from_did_web_no_change():
    diddoc = {'id': 'did:webs:example:123', 'verificationMethod': [{'controller': 'did:webs:example:123'}]}

    from dws.core.didding import from_did_web

    result = from_did_web(diddoc)

    assert result['id'] == 'did:webs:example:123'
    assert result['verificationMethod'][0]['controller'] == 'did:webs:example:123'

    unstub()


def test_parse_did_webs_no_match_raises():
    did = 'did:webs:example:123'
    with pytest.raises(ValueError) as excinfo:
        didding.parse_did_webs(did)
        assert str(excinfo.value) == f'{did} is not a valid did:web(s) DID'


def test_parse_query_string_false_param():
    query_string = '?param1=value1&param2=false'
    result = didding.parse_query_string(query_string)
    assert result == {'param1': 'value1', 'param2': False}


def test_parse_query_string_with_int_parse_as_int():
    query_string = '?param1=value1&param2=42'
    result = didding.parse_query_string(query_string)
    assert result == {'param1': 'value1', 'param2': 42}


def test_re_encode_invalid_did_webs_invalid_aid_raises():
    invalid_aid = 'EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk'
    # it is invalidly encoded because it has an unencoded colon after the domain instead of the percent-encoded %3A string.
    invl_did_invl_aid = f'did:webs:127.0.0.1:7676:{invalid_aid}'
    with pytest.raises(ValueError) as excinfo:
        didding.re_encode_invalid_did_webs(invl_did_invl_aid)
        assert str(excinfo.value) == f'{invl_did_invl_aid} is an invalid AID'


def test_re_encode_valid_did_webs_did_returns_original_did():
    valid_did = 'did:webs:example.com%3A8443:my:path:components:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4?meta=true'
    result = didding.re_encode_invalid_did_webs(valid_did)
    assert result == valid_did, f'Expected {valid_did}, but got {result}'


def test_re_encode_invalid_no_aid_did_webs_throws():
    no_aid = f'did:webs:127.0.0.1:7676'
    with pytest.raises(ValueError) as excinfo:
        didding.re_encode_invalid_did_webs(no_aid)
        assert str(excinfo.value) == f'{no_aid} is missing an AID'


@pytest.mark.parametrize(
    'invalid_did,expected_did',
    [
        # no port, no path, no query
        (
            'did:webs:example.com:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
            'did:webs:example.com:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
        ),
        # port only
        (
            'did:webs:example.com:8443:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
            'did:webs:example.com%3A8443:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
        ),
        # path only
        (
            'did:webs:example.com:my:path:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
            'did:webs:example.com:my:path:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4',
        ),
        # query only
        (
            'did:webs:example.com:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4?meta=true',
            'did:webs:example.com:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4?meta=true',
        ),
        # port + path + query
        (
            'did:webs:example.com:8443:my:path:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4?meta=true',
            'did:webs:example.com%3A8443:my:path:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4?meta=true',
        ),
    ],
)
def test_re_encode_invalid_did_webs_optional_parts(invalid_did, expected_did):
    result = didding.re_encode_invalid_did_webs(invalid_did)
    assert result == expected_did, f'Expected {expected_did}, but got {result}'


def test_re_encode_invalid_did_non_webs_raises():
    invalid_did = 'did:example:123'
    with pytest.raises(ValueError) as excinfo:
        didding.re_encode_invalid_did_webs(invalid_did)
        assert str(excinfo.value) == f'{invalid_did} is not an invalidly encoded did:web(s) DID'


def test_re_encode_invalid_did_webs_did_adds_domain_port_path_aid_query_parts():
    domain = 'example.com'
    port = '8080'
    path = 'some:path:for:did'
    aid = 'EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    query = '?meta=True'
    # it is invalidly encoded because it has an unencoded colon after the domain instead of the percent-encoded %3A string.
    invalidly_encoded_did = f'did:webs:{domain}:{port}:{path}:{aid}{query}'
    expected_did = f'did:webs:{domain}%3A{port}:{path}:{aid}{query}'
    result = didding.re_encode_invalid_did_webs(invalidly_encoded_did)
    assert result == expected_did, f'Expected {expected_did}, but got {result}'
    assert domain in result, 'Domain should be present in the result'
    assert port in result, 'Port should be present in the result'
    assert path in result, 'Path should be present in the result'
    assert aid in result, 'AID should be present in the result'
    assert query in result, 'Query should be present in the result'
    assert result.startswith('did:webs:'), "Result should start with 'did:webs:'"
    assert result.count(':') == 7, 'Result should have exactly 7 colons'
    assert '%3A' in result, "Result should contain the encoded colon '%3A' after the domain"


def test_re_encode_invalid_did_raises_for_non_did_webs_or_keri_did():
    non_did_webs_did = 'did:example:123'
    with pytest.raises(ValueError) as excinfo:
        didding.re_encode_invalid_did(non_did_webs_did)
        assert str(excinfo.value) == f'{non_did_webs_did} is not a valid did:webs or did:keri DID'


def test_re_encode_invalid_did_encodes_did_webs_did():
    invalidly_encoded_did = (
        f'did:webs:example.com:1234:my:path:components:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4?meta=true'
    )
    expected_did = f'did:webs:example.com%3A1234:my:path:components:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4?meta=true'
    result = didding.re_encode_invalid_did(invalidly_encoded_did)
    assert result == expected_did, f'Expected {expected_did}, but got {result}'


def test_re_encode_invalid_did_encodes_did_keri_did():
    did_keri_did = f'did:keri:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4'
    result = didding.re_encode_invalid_did(did_keri_did)
    assert result == did_keri_did, f'Expected {did_keri_did}, but got {result}'


def test_did_keri_regex_match_parses_query():
    aid = 'EEdpe-yqftH2_FO1-luoHvaiShK4y_E2dInrRQ2_2X5v'
    query = '?meta=true&oobi=http%3A//127.0.0.1%3A6642/oobi/EEdpe-yqftH2_FO1-luoHvaiShK4y_E2dInrRQ2_2X5v/witness/BPwwr5VkI1b7ZA2mVbzhLL47UPjsGBX4WeO6WRv6c7H-'
    did_keri_did = f'did:keri:{aid}{query}'
    matched_aid, matched_query = didding.parse_did_keri(did_keri_did)
    assert matched_aid is not None, f'Expected AID to be parsed, but got {aid}'
    assert matched_query is not None, f'Expected query to be parsed, but got {query}'
    assert matched_aid == aid, f'Expected matched AID to be {aid}, but got {matched_aid}'
    assert matched_query == query, f'Expected matched query to be {query}, but got {matched_query}'


def test_requote_did_re_encodes_improperly_url_encoded_dids():
    invalidly_encoded_did = (
        f'did%3Awebs%3Aexample.com%3A8443%3Amy%3Apath%3Acomponents%3AEKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4?meta=true'
    )
    expected_did = f'did:webs:example.com%3A8443:my:path:components:EKYLUMmNPZeEs77Zvclf0bSN5IN-mLfLpx2ySb-HDlk4?meta=true'
    requoted_did = didding.requote(invalidly_encoded_did)
    assert requoted_did == expected_did, f'Expected {expected_did}, but got {requoted_did}'


def test_extract_desg_alias_from_cred_rev_status_returns_none():
    cred = {
        'sad': {'a': {'ids': ['did:webs:foo:designated_id_2', 'designated_id_2_but_different']}},
        'status': {'et': 'rev'},
    }
    assert None == didding.extract_desg_alias_from_cred(cred), 'Expected None for revocation status'


def test_keri_resolver_with_empty_hby_creates_hby():
    hby_mock = mock(habbing.Habery)
    hby_mock.db = mock(basing.Baser)
    hby_doer_mock = mock(habbing.HaberyDoer)
    rgy_mock = mock(credentialing.Regery)
    with patch('dws.core.habs.get_habery_and_doer') as mock_get_habery_and_doer:
        mock_get_habery_and_doer.return_value = hby_mock, hby_doer_mock
        resolver = didkeri.KeriResolver(did='fake:did', name='test_resolver', base='test_base', rgy=rgy_mock)
        assert resolver.hby == hby_mock, 'Expected KeriResolver to create a new Habery instance'
        assert hby_doer_mock in resolver.doers, 'Expected KeriResolver to add HaberyDoer to its doers'


def test_designated_aliases_generation_returns_creds_when_non_local_aid():
    hby = mock(habbing.Habery)
    hby.db = mock(basing.Baser)
    hby.habs = {}
    rgy = mock(credentialing.Regery)
    rgy.reger = mock(credentialing.Reger)
    rgy.reger.issus = mock(subing.CesrDupSuber)
    rgy.reger.schms = mock(subing.CesrDupSuber)

    test_saider = coring.Saider(qb64='EN6Oh5XSD5_q2Hgu-aqpdfbVepdpYpFlgz6zvJL5b_r5')

    when(rgy.reger.issus).get(keys='test_aid').thenReturn([test_saider])
    when(rgy.reger.schms).get(keys='EN6Oh5XSD5_q2Hgu-aqpdfbVepdpYpFlgz6zvJL5b_r5').thenReturn([test_saider])
    when(rgy.reger).cloneCreds([test_saider], hby.db).thenReturn([])
    da = didding.gen_designated_aliases(hby, rgy, 'test_aid')
    assert da == [], 'Expected empty list for designated aliases when no credentials are found'


def test_generate_did_doc_with_delegator_shows_service_delegator_section():
    pass


def test_gen_delegation_service_generates_correctly():
    seal_evt = dict(i='delegate_aid', s='0', d='delegate_aid')
    mockSealSerder = mock(serdering.SerderKERI)
    mockSealSerder.sad = {'a': [{'d': 'delegator_aid'}]}

    hby = mock(habbing.Habery)
    hby.db = mock(basing.Baser)
    hby.db.roobi = mock(koming.Komer)
    when(hby.db).fetchLastSealingEventByEventSeal(pre='delegator_aid', seal=seal_evt).thenReturn(mockSealSerder)
    oobi = 'http://example.com/oobi/delegate_aid'
    obr = basing.OobiRecord(cid='delegator_aid')
    when(hby.db.roobi).getItemIter().thenReturn([((oobi,), obr)])

    assert didding.gen_delegation_service(hby=hby, pre='delegate_aid', delpre='delegator_aid')


def test_gen_delegation_service_no_oobi_returns_none():
    seal_evt = dict(i='delegate_aid', s='0', d='delegate_aid')
    mockSealSerder = mock(serdering.SerderKERI)
    mockSealSerder.sad = {'a': [{'d': 'delegator_aid'}]}

    hby = mock(habbing.Habery)
    hby.db = mock(basing.Baser)
    hby.db.roobi = mock(koming.Komer)
    when(hby.db).fetchLastSealingEventByEventSeal(pre='delegator_aid', seal=seal_evt).thenReturn(mockSealSerder)
    when(hby.db.roobi).getItemIter().thenReturn([])

    assert didding.gen_delegation_service(hby=hby, pre='delegate_aid', delpre='delegator_aid') == []


def test_gen_service_endpoints_when_del_serv_end_is_none_does_not_include_delegator_service():
    aid = 'delegate_aid'
    seal_evt = dict(i='delegate_aid', s='0', d='delegate_aid')
    mockSealSerder = mock(serdering.SerderKERI)
    mockSealSerder.sad = {'a': [{'d': 'delegator_aid'}]}

    hby = mock(habbing.Habery)
    hby.db = mock(basing.Baser)
    hby.db.roobi = mock(koming.Komer)
    when(hby.db).fetchLastSealingEventByEventSeal(pre='delegator_aid', seal=seal_evt).thenReturn(mockSealSerder)
    when(hby.db.roobi).getItemIter().thenReturn([])

    hab = mock(habbing.Hab)
    kever = mock(eventing.Kever)
    when(hab).fetchRoleUrls(cid=aid).thenReturn(role_urls_fixture())
    when(hab).fetchWitnessUrls(cid=aid).thenReturn(witness_urls_fixture())

    serv_endpoints = didding.gen_service_endpoints(hby=hby, hab=hab, kever=kever, aid='delegate_aid')
    assert len(serv_endpoints) > 0, 'Service endpoints should not be empty'
    assert 'DelegatorOOBI' not in [service['type'] for service in serv_endpoints], 'Delegator service should not be included'



import logging

from dws import log_name, ogler, set_log_level


def test_set_log_level():
    logger = ogler.getLogger(log_name)
    assert logger.level == logging.INFO  # Default log level
    set_log_level('debug', logger)
    assert logger.level == logging.DEBUG



import json
from typing import List

from hio.base import Doer, doing
from keri.app import oobiing
from keri.app.configing import Configer
from keri.app.habbing import Habery
from keri.vdr.credentialing import Regery

from dws import log_name, ogler
from dws.core import artifacting, habs

logger = ogler.getLogger(log_name)


class DIDArtifactGenerator(doing.DoDoer):
    """
    Generates a did:webs DID document and the associated CESR stream for the {AID}.json and keri.cesr files.
    - {AID}.json contains the DID document
    - keri.cesr contains the CESR event stream for the KELs, TELs, and ACDCs associated with the DID.
    """

    def __init__(
        self,
        name: str,
        base: str | None,
        bran: str | None,
        did: str,
        meta: bool = False,
        output_dir: str = '.',
        verbose: bool = False,
        config_dir: str | None = None,
        config_file: str | None = None,
        cf: Configer | None = None,
        hby: Habery | None = None,
        hby_doer: Doer | None = None,
        regery: Regery | None = None,
    ):
        """
        Initializes the did:webs DID file generator.

        Parameters:
            name (str): Name of the controller keystore (Habery) to use for generating the DID document.
            base (str): Base path (namespace for local file tree) for the KERI keystore (Habery) to use for generating the DID document.
            bran (str): Passcode for the controller of the local KERI keystore.
            did (str): The did:webs DID showing the domain and AID to generate the DID document and CESR stream for.
            meta (bool): Whether to include metadata in the DID document generation. Defaults to False.
            output_dir (str): Directory to output the generated files. Default is current directory.
            verbose (bool): Whether to print the generated DID artifacts at the command line. Defaults to False
            config_dir (str): Directory override for configuration data. Defaults to None.
            config_file (str): Configuration filename override. Defaults to None.
            cf (Configer): Optional Configer instance to use for configuration. If None, it will be created based on the provided parameters.
            hby (Habery): Optional Habery instance to use. If None, it will be created based on the provided parameters.
            hby_doer (Doer): Optional Doer instance for the Habery. If None, it will be created based on the provided parameters.
            regery (Regery): Optional Regery instance to use. If None, it will be created based on the provided Habery.
        """
        self.name: str = name
        self.base: str = base
        self.bran: str = bran
        cf = cf if cf else habs.get_habery_configer(name=config_file, base=base, head_dir_path=config_dir)
        if hby is None and hby_doer is None:
            hby, hby_doer = habs.get_habery_and_doer(name, base, bran, cf)
        self.hby: Habery = hby
        self.rgy: Regery = regery if regery else Regery(hby=self.hby, name=self.hby.name, base=self.hby.base)
        oobiery = oobiing.Oobiery(hby=self.hby)
        self.did: str = did
        self.meta: bool = meta
        self.verbose: bool = verbose
        self.output_dir: str = output_dir

        self.did_json = {}
        self.keri_cesr = bytearray()

        self.toRemove: List[Doer] = [hby_doer] + oobiery.doers
        doers = list(self.toRemove)
        super(DIDArtifactGenerator, self).__init__(doers=doers)

    def recur(self, tock=0.0, **opts):
        """DoDoer lifecycle function that calls the underlying DID generation function. Runs once"""
        self.generate()
        return True  # run once and then stop

    def generate(self):
        """Drive did:webs did.json and keri.cesr generation"""
        logger.debug(f'\nGenerate DID doc for: {self.did}\nand metadata        : {self.meta}')
        did_json, keri_cesr = artifacting.generate_artifacts(self.hby, self.rgy, self.did, self.meta, self.output_dir)

        if self.verbose:
            print(f'keri.cesr:\n{keri_cesr.decode()}\n')
            print(f'did.json:\n{json.dumps(did_json, indent=2)}')
        self.did_json = did_json
        self.keri_cesr = keri_cesr
        self.remove(self.toRemove)
        return True



import json
import os

from hio.core import http
from keri import kering
from keri.app import habbing, signing
from keri.app.habbing import Hab, Habery
from keri.core.serdering import SerderACDC
from keri.vdr import credentialing, viring
from keri.vdr.credentialing import Regery
from keri.vdr.viring import Reger

from dws import UnknownAID, log_name, ogler
from dws.core import didding, ends, resolving, webbing
from dws.core.didding import DES_ALIASES_SCHEMA

logger = ogler.getLogger(log_name)


def gen_kel_cesr(hab: habbing.Hab, pre: str) -> bytearray:
    """Return a bytearray of the CESR stream of all KEL events for a given prefix."""
    return hab.replay(pre=pre)


def make_keri_cesr_path(output_dir: str, aid: str):
    """Create keri.cesr enclosing dir, and any intermediate dirs, if not existing"""
    kc_dir_path = os.path.join(output_dir, aid)
    if not os.path.exists(kc_dir_path):
        logger.debug(f'Creating directory for KERI CESR events: {kc_dir_path}')
        os.makedirs(kc_dir_path)
    return os.path.join(kc_dir_path)


def write_keri_cesr_file(output_dir: str, aid: str, keri_cesr: bytearray):
    """Write the keri.cesr file to output path, making any enclosing directories"""
    kc_file_path = make_keri_cesr_path(output_dir, aid)
    kc_file_path = os.path.join(kc_file_path, ends.KERI_CESR)
    with open(kc_file_path, 'w') as kcf:
        tmsg = keri_cesr.decode('utf-8')
        logger.debug(f'Writing CESR events to {kc_file_path}: \n{tmsg}')
        kcf.write(tmsg)


def get_self_issued_acdcs(aid: str, reger: credentialing.Reger, schema: str = didding.DES_ALIASES_SCHEMA):
    """Get self issued ACDCs filtered by schema"""
    creds_issued = reger.issus.get(keys=aid)
    creds_by_schema = reger.schms.get(keys=schema.encode('utf-8'))

    # self-attested, there is no issuee, and schema is designated aliases
    return [
        cred_issued
        for cred_issued in creds_issued
        if cred_issued.qb64 in [cred_by_schm.qb64 for cred_by_schm in creds_by_schema]
    ]


def gen_tel_cesr(reger: viring.Reger, evt_pre: str) -> bytearray:
    """Get the CESR stream of TEL events for a given registry."""
    msgs = bytearray()
    for msg in reger.clonePreIter(pre=evt_pre):
        msgs.extend(msg)
    return msgs


def gen_acdc_cesr(hab: Hab, reger: Reger, creder: SerderACDC) -> bytearray:
    """
    Add the CESR stream of the self attestation ACDCs for the given AID including signatures
    and their anchors to their source KELs.
    """
    arr = bytearray()
    (prefixer, seqner, saider) = reger.cancs.get(keys=(creder.said,))
    arr.extend(bytearray(signing.serialize(creder, prefixer, seqner, saider)))
    return arr


def gen_des_aliases_cesr(hab: Hab, reger: Reger, aid: str, schema: str = DES_ALIASES_SCHEMA) -> bytearray:
    """
    Select a specific ACDC from the local registry (Regery), if it exists, to generate the
    CESR stream
    Args:
        hab: The local Hab to use for generating the CESR stream
        aid: AID prefix to retrieve the ACDC for
        reger: The Regery to use for retrieving the ACDC
        schema: the schema to use to select the target ACDC from the local registry

    Returns:
        bytearray: CESR stream of locally stored ACDC events for the specified AID and schema
    """
    # self-attested, there is no issuee, and schema is designated aliases
    local_creds = get_self_issued_acdcs(aid, reger, schema)

    cesr_bytes = bytearray()
    for cred in local_creds:
        creder, *_ = reger.cloneCred(said=cred.qb64)
        cesr_bytes.extend(add_cred_cesr_bytes(creder, hab, reger))
    return cesr_bytes


def add_cred_cesr_bytes(creder: SerderACDC, hab: Hab, reger: Reger) -> bytearray:
    """Add one ACDC credential bytes to the CESR stream"""
    creder_bytes = bytearray()
    creder_bytes.extend(gen_tel_cesr(reger, creder.regi))
    creder_bytes.extend(gen_tel_cesr(reger, creder.said))
    creder_bytes.extend(gen_acdc_cesr(hab, reger, creder))
    return creder_bytes


def get_witness_loc_scheme_bytes(hab: habbing.Hab, wit_prefixes: list[str], scheme: str = '') -> bytearray:
    """Gets the witness location scheme records from the local hab database as a bytearray."""
    msgs = bytearray()
    for eid in wit_prefixes:
        loc_scheme_msg = hab.loadLocScheme(eid=eid, scheme=scheme)
        logger.debug(f'Found witness location scheme message for eid {eid}: {loc_scheme_msg}')
        msgs.extend(loc_scheme_msg if loc_scheme_msg else bytearray())
        end_role_msg = hab.loadEndRole(cid=eid, eid=eid, role=kering.Roles.controller)
        logger.debug(f'Found witness endpoint role message for eid {eid}: {end_role_msg}')
        msgs.extend(end_role_msg if end_role_msg else bytearray())
    return msgs


def get_agent_loc_scheme_bytes(hab: habbing.Hab, aid: str, scheme: str = '') -> bytearray:
    """Gets the agent location scheme records from the local hab database as a bytearray."""
    msgs = bytearray()
    for (_, erole, eid), _ in hab.db.ends.getItemIter(keys=(aid, kering.Roles.agent)):
        # Get the location scheme message for the agent
        loc_scheme_msg = hab.loadLocScheme(eid=eid, scheme=scheme)
        logger.debug(f'Found agent location scheme message for eid {eid}: {loc_scheme_msg}')
        msgs.extend(loc_scheme_msg if loc_scheme_msg else bytearray())
        # Get the endpoint role message for the agent
        end_role_msg = hab.loadEndRole(cid=aid, eid=eid, role=erole)
        logger.debug(f'Found agent endpoint role message for eid {eid}: {end_role_msg}')
        msgs.extend(end_role_msg if end_role_msg else bytearray())
    return msgs


def get_mailbox_loc_scheme_endrole_bytes(hab: habbing.Hab, aid: str, scheme: str = '') -> bytearray:
    """Gets the mailbox location scheme and endpoint role records from the local hab database as a bytearray."""
    msgs = bytearray()
    for (_, erole, eid), _ in hab.db.ends.getItemIter(keys=(aid, kering.Roles.mailbox)):
        # Get the location scheme message for the mailbox
        loc_scheme_msg = hab.loadLocScheme(eid=eid)
        logger.debug(f'Found mailbox location scheme message for eid {eid}: {loc_scheme_msg}')
        msgs.extend(loc_scheme_msg if loc_scheme_msg else bytearray())
        # Get the endpoint role message for the mailbox
        end_role_msg = hab.loadEndRole(cid=aid, eid=eid, role=erole)
        logger.debug(f'Found mailbox endpoint role message for eid {eid}: {end_role_msg}')
        msgs.extend(end_role_msg if end_role_msg else bytearray())
    return msgs


def gen_loc_schemes_cesr(hab: habbing.Hab, aid: str, role: str = None, scheme='') -> bytearray:
    """
    Generates a CESR stream of all location scheme record reply 'rpy' messages for a given AID based on
    the witness location scheme and endpoint role records in the local Hab's database.

    Returns:
        bytearray: CESR stream of location scheme and endpoint role records for the given AID and role.

    Parameters:
        hab (habbing.Hab): The local Hab to use for generating the CESR stream.
        aid (str): The AID prefix to retrieve the location schemes for.
        role (str): The role of the endpoint, e.g., witness, agent, mailbox. Defaults to None.
        scheme (str): The scheme to filter the location schemes by. Defaults to an empty string.
    """
    kever = hab.kevers[aid]
    msgs = bytearray()
    # Get witness location schemes and endpoint roles
    if not role or role == kering.Roles.witness:
        msgs.extend(get_witness_loc_scheme_bytes(hab, kever.wits, scheme=scheme))
    # Get agent and mailbox location schemes and endpoint roles
    if not role or role == kering.Roles.agent:  # in preparation for working with KERIA agents
        msgs.extend(get_agent_loc_scheme_bytes(hab, aid, scheme=scheme))
    # Get mailbox location schemes and endpoint roles
    if not role or role == kering.Roles.mailbox:
        msgs.extend(get_mailbox_loc_scheme_endrole_bytes(hab, aid, scheme=scheme))
    return msgs


def make_did_json_path(output_dir: str, aid: str):
    """Create the directory (and any intermediate directories in the given path) if it doesn't already exist"""
    dd_dir_path = os.path.join(output_dir, aid)
    if not os.path.exists(dd_dir_path):
        os.makedirs(dd_dir_path)
    return dd_dir_path


def write_did_json_file(dd_dir_path: str, diddoc: dict, meta: bool = False):
    """save did.json to a file at output_dir/{aid}/{AID}.json"""
    dd_file_path = os.path.join(dd_dir_path, f'{ends.DID_JSON}')
    with open(dd_file_path, 'w') as ddf:
        json.dump(didding.to_did_web(diddoc, meta), ddf)


def generate_artifacts(hby: Habery, rgy: Regery, did: str, meta: bool = False, output_dir: str = '.'):
    domain, port, path, aid, query = didding.parse_did_webs(did)

    # generate did doc
    try:
        did_json = didding.generate_did_doc(hby, rgy, did=did, aid=aid, meta=meta)
    except UnknownAID as e:
        logger.error(f'Failed to generate DID document for {did}: {e}')
        raise e
    # Create the directory (and any intermediate directories in the given path) if it doesn't already exist
    dd_dir_path = make_did_json_path(output_dir, aid)
    write_did_json_file(dd_dir_path, did_json, meta)

    logger.info(f'Generating CESR event stream data from local Habery keystore')
    hab = hby.habs[aid]
    reger = rgy.reger
    keri_cesr = bytearray()
    keri_cesr.extend(gen_kel_cesr(hab, aid))  # add KEL CESR stream
    keri_cesr.extend(gen_loc_schemes_cesr(hab, aid))  # add witness location schemes
    keri_cesr.extend(gen_des_aliases_cesr(hab, reger, aid))  # add designated aliases TELs and ACDCs
    write_keri_cesr_file(output_dir, aid, keri_cesr)

    return did_json, keri_cesr


def dyn_artifact_svr_doers(
    hby, rgy, alias: str, http_port, did_path=None, meta=False, keypath=None, certpath=None, cafilepath=None
):
    """
    Parameters:
        hby (habbing.Habery): identifier database environment
        rgy (credentialing.Regery): Doer for the identifier database environment
        alias (str): alias for the KERI AID to dynamically generate and serve did:webs and did:keri artifacts for.
        http_port (int): external port to listen on for HTTP messages
        meta (bool): whether to include metadata in the DID document, default is False
        did_path (str): path segment of the did:webs URL to host the did:webs artifacts on, disabled if None
        keypath (str | None): path to the TLS private key file, default is None (disabled)
        certpath (str | None): path to the TLS certificate file, default is None (disabled)
        cafilepath (str | None): path to the CA certificate file, default is None (disabled)
    Returns:
        list: list of Doers to run in the Tymist
    """
    doers = []
    app = resolving.falcon_app()
    webbing.load_endpoints(app, hby=hby, rgy=rgy, did_path=did_path, meta=meta)
    server = resolving.tls_falcon_server(app, http_port, keypath, certpath, cafilepath)
    http_server_doer = http.ServerDoer(server=server)
    doers.extend([http_server_doer])
    return doers



from keri.db import subing
from keri.vdr import credentialing
from mockito import mock, verify, when

from dws.core import didding
from dws.core.webbing import load_endpoints


def test_setup():
    app = mock()
    hby = mock()
    hby.name = 'test_hab'
    hby.base = 'test_base'

    rgy = mock(credentialing.Regery)
    rgy.reger = mock(credentialing.Reger)
    rgy.reger.issus = mock(subing.CesrDupSuber)
    rgy.reger.schms = mock(subing.CesrDupSuber)
    when(rgy.reger.issus).get(keys='test_aid').thenReturn([])
    when(rgy.reger.schms).get(keys=didding.DES_ALIASES_SCHEMA).thenReturn([])

    load_endpoints(app, hby, rgy)

    verify(app, times=1).add_route('/{aid}/did.json', any)
    verify(app, times=1).add_route('/{aid}/keri.cesr', any)



# -*- encoding: utf-8 -*-
"""
dws.app.cli.commands.did.webs.generate module

"""

import argparse

from hio.base import doing

from dws import log_name, ogler, set_log_level
from dws.core.generating import DIDArtifactGenerator

parser = argparse.ArgumentParser(description='Generate a did:webs DID document and KEL, TEL, and ACDC CESR stream file')
parser.set_defaults(handler=lambda args: handler(args), transferable=True)
parser.add_argument('-n', '--name', action='store', required=True, help='Name of controller.')
parser.add_argument(
    '-b', '--base', required=False, default='', help='additional optional prefix to file location of KERI keystore'
)
# passcode => bran
parser.add_argument(
    '-p', '--passcode', dest='bran', default=None, help='22 character encryption passcode for keystore (is not saved)'
)
parser.add_argument('--config-dir', '-c', dest='config_dir', help='directory override for configuration data', default=None)
parser.add_argument('--config-file', dest='config_file', action='store', default=None, help='configuration filename override')
parser.add_argument(
    '--output-dir',
    required=False,
    default='.',
    help='Directory to output the generated files. Default is current directory.',
)
parser.add_argument(
    '-m',
    '--meta',
    action='store_true',
    required=False,
    default=False,
    help='Whether to include metadata (True), or only return the DID document (False)',
)
parser.add_argument('-d', '--did', required=True, help='DID to generate (did:webs method)')
parser.add_argument(
    '-v',
    '--verbose',
    action='store_true',
    required=False,
    default=False,
    help='Show the verbose output of DID generation artifacts.',
)
parser.add_argument(
    '--loglevel',
    action='store',
    required=False,
    default='CRITICAL',
    help='Set log level to DEBUG | INFO | WARNING | ERROR | CRITICAL. Default is CRITICAL',
)

logger = ogler.getLogger(log_name)


def handler(args: argparse.Namespace) -> list[doing.Doer]:
    """
    Return a list of HIO Doers that perform did:webs artifact generation for the DID document and
    keri.cesr CESR stream and then shut down.
    """
    set_log_level(args.loglevel, logger)
    return [
        DIDArtifactGenerator(
            name=args.name,
            base=args.base,
            bran=args.bran,
            config_dir=args.config_dir,
            config_file=args.config_file,
            did=args.did,
            meta=args.meta,
            output_dir=args.output_dir,
            verbose=args.verbose,
        )
    ]



# -*- encoding: utf-8 -*-
"""
dws.app.cli.commands.did.webs.resolver-service module

"""

import argparse
from typing import List

from hio.base import doing
from keri.app import oobiing
from keri.vdr import credentialing

from dws import log_name, ogler, set_log_level
from dws.core import habs, resolving

parser = argparse.ArgumentParser(description='Expose did:webs resolver as an HTTP web service')
parser.set_defaults(handler=lambda args: launch(args), transferable=True)
parser.add_argument(
    '-p',
    '--http',
    action='store',
    default=7677,
    help='Port on which to listen for did:webs resolution requests.  Defaults to 7677',
)
parser.add_argument(
    '-d',
    '--did-path',
    action='store',
    default='',
    required=False,
    help="did:webs path segment in URL format between {host}%3A{port} and {aid}. Example: 'somepath/somesubpath'",
)
parser.add_argument('-n', '--name', action='store', required=True, help='Name of controller.')
parser.add_argument(
    '-b', '--base', required=False, default='', help='additional optional prefix to file location of KERI keystore'
)
# passcode => bran
parser.add_argument(
    '--passcode', dest='bran', default=None, help='22 character encryption passcode for keystore (is not saved)'
)
parser.add_argument('-c', '--config-dir', dest='config_dir', default=None, help='directory override for configuration data')
parser.add_argument('--config-file', dest='config_file', action='store', default=None, help='configuration filename override')
parser.add_argument(
    '--static-files-dir',
    dest='static_files_dir',
    action='store',
    default=None,
    help='static files directory to use for serving the did.json and keri.cesr files. Default is "static"',
)
parser.add_argument('--keypath', action='store', required=False, default=None)
parser.add_argument('--certpath', action='store', required=False, default=None)
parser.add_argument('--cafilepath', action='store', required=False, default=None)
parser.add_argument(
    '--loglevel',
    action='store',
    required=False,
    default='CRITICAL',
    help='Set log level to DEBUG | INFO | WARNING | ERROR | CRITICAL. Default is CRITICAL',
)

logger = ogler.getLogger(log_name)


def launch(args, expire=0.0):
    """
    Launches a Falcon webserver listening on /1.0/identifiers/{did} for did:webs resolution requests
    as a set of Doers
    """
    set_log_level(args.loglevel, logger)
    http_port = args.http
    try:
        http_port = int(http_port)
    except ValueError:
        logger.error(f'Invalid port number: {http_port}. Must be an integer.')
        raise

    doers = create_did_webs_doers(
        name=args.name,
        base=args.base,
        bran=args.bran,
        config_file=args.config_file,
        config_dir=args.config_dir,
        static_files_dir=args.static_files_dir,
        did_path=args.did_path,
        http_port=http_port,
        keypath=args.keypath,
        certpath=args.certpath,
        cafilepath=args.cafilepath,
    )
    logger.info(f'Launched did:webs resolver on {http_port}')
    return doers


def create_did_webs_doers(
    name: str,
    base: str,
    bran: str,
    config_file: str,
    config_dir: str,
    static_files_dir: str,
    did_path: str,
    http_port: int,
    keypath: str,
    certpath: str,
    cafilepath: str,
) -> List[doing.Doer]:
    cf = habs.get_habery_configer(name=config_file, base=base, head_dir_path=config_dir)
    hby, hby_doer = habs.get_habery_and_doer(name, base, bran, cf)
    rgy = credentialing.Regery(hby=hby, name=hby.name, base=hby.base, temp=hby.temp)
    oobiery = oobiing.Oobiery(hby=hby)

    doers = [hby_doer] + oobiery.doers
    doers += resolving.setup_resolver(
        hby,
        rgy,
        oobiery,
        http_port=http_port,
        static_files_dir=static_files_dir,
        did_path=did_path,
        keypath=keypath,
        certpath=certpath,
        cafilepath=cafilepath,
    )
    return doers



from hio.help import hicting
from keri import kering
from keri.app import configing, habbing, keeping
from keri.app.cli.common import existing
from keri.core import eventing
from keri.db import basing

from dws import log_name, ogler

logger = ogler.getLogger(log_name)


def get_habery_configer(name: str | None, base: str | None, head_dir_path: str | None, temp: bool = False):
    """Get the Configer for the Habery if name provide otherwise return None."""
    if name is not None:
        return configing.Configer(name=name, base=base, headDirPath=head_dir_path, temp=temp, reopen=True, clear=False)
    return None


def get_auth_encryption_aid(name: str, base: str, temp: bool = False):
    """Get the Authentication and Encryption Identifier (AEID) from the Keeper."""
    ks = keeping.Keeper(name=name, base=base, temp=temp, reopen=True)
    aeid = ks.gbls.get('aeid')
    ks.close()  # to avoid LMDB reader table locks
    return aeid


def get_habery_and_doer(
    name: str | None, base: str | None, bran: str | None, cf: configing.Configer = None, temp: bool = False
) -> (habbing.Habery, habbing.HaberyDoer):
    """Get the Habery and its Doer respecting any existing AEID."""
    aeid = get_auth_encryption_aid(name, base, temp)
    if aeid is None:
        hby = habbing.Habery(name=name, base=base, bran=bran, cf=cf, temp=temp)
    else:
        hby = existing.setupHby(name=name, base=base, bran=bran, cf=cf, temp=temp)
    return hby, habbing.HaberyDoer(habery=hby)


def fetch_urls(baser: basing.Baser, eid: str, scheme: str = '') -> hicting.Mict:
    """
    Returns:
        hicting.Mict: urls keyed by scheme for given endpoint identifier (eid). Assumes that user
            independently verifies that the eid is allowed for a given cid and role.
            If url is empty then does not return.

    Parameters:
        baser (basing.Baser): The Baser instance to fetch URLs from.
        eid (str): The endpoint identifier (eid) to fetch URLs for.
        scheme (str): The scheme to filter URLs by. Defaults to an empty string, which means no filtering.
    """
    urls = []
    for keys, loc in baser.locs.getItemIter(keys=(eid, scheme)):
        logger.debug(f'Fetched URL: {loc.url} for eid: {eid}, scheme: {scheme} with keys: {keys}')
        if loc.url:
            urls.append((keys[1], loc.url))
    return hicting.Mict(urls)


def get_role_urls(baser: basing.Baser, kever: eventing.Kever, scheme: str = ''):
    """
    Gets all witness role URLs in a given database for the witnesses in a Kever, filterable by scheme.

    A subset of the habbing.BaseHab.fetchRoleUrls method from KERIpy.

    Parameters:
        baser (basing.Baser): The Baser instance to fetch URLs from.
        kever (eventing.Kever): The Kever instance containing the witnesses.
        scheme (str): The scheme to filter URLs by. Defaults to an empty string, which means no filtering.
    """
    rurls = hicting.Mict()
    for eid in kever.wits:
        surls = fetch_urls(baser, eid, scheme=scheme)
        if surls:
            rurls.add(kering.Roles.witness, hicting.Mict([(eid, surls)]))
    return rurls



import asyncio

from didcomm.common.resolvers import ResolversConfig
from didcomm.message import Message
from didcomm.pack_encrypted import PackEncryptedConfig, pack_encrypted
from didcomm.unpack import unpack
from utils import DidKeriResolver, SecretsResolverInMemory, createKeriDid, validateLongDid

alice = createKeriDid()
print("Alice's DID:", alice['did'])
print("Alice's Long DID:", alice['long_did'], '\n')

bob = createKeriDid()
print("Bob's DID:", bob['did'])
print("Bob's Long DID:", bob['long_did'], '\n')

print(
    'Compare length with Peer DID: did:peer:2.Vz6MkfiT7uT2EQnuNhGP2xxNsh2v7seoV3fzZWRpydJnF3K6z.Ez6LSfpsWUKP7g4LGTpXFPWZzThQtG1VuQqxtToKYtkXdRdc8.SeyJpZCI6IiNzZXJ2aWNlLTEiLCJ0IjoiZG0iLCJzIjoiZGlkOnBlZXI6Mi5FejZMU2U3U3FCUmJmWEdFNXJpVEZYYU5zZ2dWUFB6YThtNXdyR21ON0dRY0d5dHI2LlZ6Nk1ra3dmNWhWcUJVYXhFU3YxVHhra0pIQWNKRldnTXROY1g2Q1pyRkZpZllNQTkuU2V5SnBaQ0k2SW01bGR5MXBaQ0lzSW5RaU9pSmtiU0lzSW5NaU9pSm9kSFJ3Y3pvdkwyMWxaR2xoZEc5eUxuSnZiM1J6YVdRdVkyeHZkV1FpTENKaElqcGJJbVJwWkdOdmJXMHZkaklpWFgwIn',
    '\n',
)

print("Alice's long DID validation:", validateLongDid(alice['long_did']))
print("Bob's long DID validation:", validateLongDid(bob['long_did']), '\n')

store = {alice['did']: alice, bob['did']: bob}

secrets_resolver = SecretsResolverInMemory(store)
did_resolver = DidKeriResolver(store)

# Alice creates a basic message
alice_message = Message(
    id='123',
    type='https://didcomm.org/basicmessage/2.0/message',
    body={'content': 'Hello Bob!'},
)
print('1-Alice creates a basic message:', alice_message.body, '\n')

# Alice encrypts the message for Bob
alice_message_packed = asyncio.run(
    pack_encrypted(
        resolvers_config=ResolversConfig(secrets_resolver=secrets_resolver, did_resolver=did_resolver),
        message=alice_message,
        frm=alice['long_did'],
        to=bob['long_did'],
        sign_frm=None,
        pack_config=PackEncryptedConfig(protect_sender_id=False),
    )
)
print('2-Alice encrypts the message for Bob:')
print(alice_message_packed.packed_msg, '\n')

# Bob decrypts the message
bob_message_unpacked = asyncio.run(
    unpack(
        resolvers_config=ResolversConfig(secrets_resolver=secrets_resolver, did_resolver=did_resolver),
        packed_msg=alice_message_packed.packed_msg,
    )
)
print('3-Bob decrypts the message:', bob_message_unpacked.message.body, '\n')

# Bob creates a basic message response
bob_message = Message(
    id='124',
    type='https://didcomm.org/basicmessage/2.0/message',
    body={'content': 'Hello Alice!'},
)
print('4-Bob creates a response using short DIDs:', bob_message.body, '\n')

# Bob encrypts the message for Alice
bob_message_packed = asyncio.run(
    pack_encrypted(
        resolvers_config=ResolversConfig(secrets_resolver=secrets_resolver, did_resolver=did_resolver),
        message=bob_message,
        frm=bob['did'],
        to=alice['did'],
        sign_frm=bob['did'],
        pack_config=PackEncryptedConfig(protect_sender_id=False),
    )
)
print('5-Bob encrypts and sign the message for Alice:')
print(bob_message_packed.packed_msg, '\n')

# Alice decrypts the message
alice_message_unpacked = asyncio.run(
    unpack(
        resolvers_config=ResolversConfig(secrets_resolver=secrets_resolver, did_resolver=did_resolver),
        packed_msg=bob_message_packed.packed_msg,
    )
)
print('6-Alice decrypts the message with short DIDs:', alice_message_unpacked.message.body, '\n')

print(asyncio.run(did_resolver.resolve(alice['did'])))



from keri.app import habbing


def get_resolved_oobi(hby: habbing.Habery, pre: str) -> str | None:
    for (oobi,), obr in hby.db.roobi.getItemIter():
        if obr.cid == pre:
            return oobi
    return None
